#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This module contains functionality to export alignments
for visualization with pianoprecision:
https://github.com/yucongj/piano-precision
and sonic visualizer:
https://www.sonicvisualiser.org/
"""

import numpy as np
import partitura as pt
from fractions import Fraction
import csv
from pathlib import Path
import numpy.lib.recfunctions as rfn
from partitura.utils.generic import interp1d


def save_piano_precision_csv(performance, spart, alignment, out="scorealignment.csv"):
    """
    save alignment for pianoprecision

    Parameters
    ----------

    performance : object
        a performance object
    spart: object
        a score part object
    alignment: list
        note alignment list of dicts

    """

    pp_list = convert_alignment_to_list(alignment, spart, performance)
    export_piano_precision_to_csv(pp_list, out)


def export_piano_precision_to_csv(pp_list, out="scorealignment.csv"):
    with open(out, mode="w", newline="") as file:
        writer = csv.writer(file)
        writer.writerow(["LABEL", "TIME"])
        for integer, frac, flt in pp_list:
            mixed = f"{integer}+{frac.numerator}/{frac.denominator}"
            writer.writerow([mixed, f"{flt:.3f}"])


def convert_alignment_to_list(alignment, spart, performance):
    score_note_array = spart.note_array()
    performance_note_array = performance.note_array()
    beat_map = spart.beat_map
    (
        _,
        stime_to_ptime_map,
    ) = pt.musicanalysis.performance_codec.get_time_maps_from_alignment(
        performance_note_array, score_note_array, alignment
    )

    measures = np.array(list(spart.iter_all(pt.score.Measure)))
    measure_starts_divs = np.array([m.start.t for m in measures])
    # measure_starts_beats = beat_map(measure_starts_divs)
    measure_sorting_idx = measure_starts_divs.argsort()
    measure_starts_divs = measure_starts_divs[measure_sorting_idx]
    measures = measures[measure_sorting_idx]

    # start_measure_num = 0 if measure_starts_beats.min() < 0 else 1
    start_measure_num = 1
    measure_starts = np.column_stack(
        (
            np.arange(start_measure_num, start_measure_num + len(measure_starts_divs)),
            measure_starts_divs,
            measures,
        )
    )
    used_note_onsets_divs = set()
    line_tuples = list()
    for mnum, msd, m in measure_starts:
        snotes = spart.iter_all(pt.score.Note, m.start, m.end, include_subclasses=True)
        for snote in snotes:
            onset_divs = snote.start.t
            if onset_divs in used_note_onsets_divs:
                continue
            else:
                onset_beats = beat_map(onset_divs)
                divs_per_quarter = int(spart.quarter_duration_map(onset_divs))
                fraction_offset_from_measure_begin = Fraction(
                    int(onset_divs - msd), (divs_per_quarter * 4)
                )
                performance_onset_sec = stime_to_ptime_map(onset_beats)
                line_tuple = (
                    int(mnum),
                    fraction_offset_from_measure_begin,
                    performance_onset_sec,
                )
                line_tuples.append(line_tuple)
                used_note_onsets_divs.add(onset_divs)

    return line_tuples


def save_notes_for_sonic_visualiser(notes_array, output_filename):
    """
    Saves a structured numpy array to a Sonic Visualiser-compatible note file.

    Parameters
    ----------
    alignment: list
        note alignment list of dicts
    performance : object
        a performance object
    spart: object
        a score part object
    """
    with open(output_filename, "w") as f:
        for note in notes_array:
            onset = float(note["onset"])
            duration = float(note["duration"])
            pitch = note["pitch"]
            # Combine ID and pitch or use just pitch if you prefer
            label = f"{pitch}"  # Or f"{note['id']}:{pitch}" if ID is meaningful
            f.write(f"{onset:.6f}\t{duration:.6f}\t{label}\n")


def save_notes_for_sonic_visualiser(note_array, out="notes.csv"):
    """
    Saves a performance note array to a Sonic Visualiser note file.

    Parameters
    ----------
    note_array: ndarray
        performance note array
    out : str
        csv filename to store the notes
    """
    with open(out, "w") as f:
        f.write("TIME,VALUE,DURATION,LEVEL,LABEL\n")
        for note in note_array:
            onset = float(note["onset_sec"])
            duration = float(note["duration_sec"])
            pitch = note["pitch"]
            label = note["id"]
            f.write(f"{onset:.9f},{pitch},{duration:.9f},0.8,{label}\n")


def save_attribute_for_sonic_visualiser_instants(
    note_array, attribute_name, out="instants.csv"
):
    """
    Saves a note array attribute array to a Sonic Visualiser instants file.

    Parameters
    ----------
    note_array: ndarray
        note array
    attribute_name: str
        name of the attribute to use as label
    out : str
        csv filename to store the instants
    """
    # # get unique instants
    # _, unique_idx = np.unique(note_array[attribute_name], return_index = True)
    with open(out, "w") as f:
        f.write("TIME,LABEL\n")
        for note in note_array:
            onset = float(note["onset_sec"])
            label = str(note[attribute_name])
            f.write(f"{onset:.9f},{label}\n")


def save_attribute_for_sonic_visualiser_time_values(
    note_array, attribute_name, out="time_values.csv", set_range=False
):
    """
    Saves a note array attribute array to a Sonic Visualiser time values file.

    Parameters
    ----------
    note_array: ndarray
        note array
    attribute_name: str
        name of the attribute to use for values
    out : str
        csv filename to store the instants
    """

    with open(out, "w") as f:
        f.write("TIME,VALUE,LABEL\n")
        for note in note_array:
            onset = float(note["onset_sec"])
            tvalue = float(note[attribute_name])
            f.write(f"{onset:.9f},{tvalue:.9f},{attribute_name}\n")
        if set_range:
            f.write(f"{onset+1:.9f},{set_range[0]:.9f},min value\n")
            f.write(f"{onset+2:.9f},{set_range[1]:.9f},max value\n")


def compute_snote_pnote_array(performance, score_part, alignment):
    """
    Saves a note array attribute array to a Sonic Visualiser time values file.

    Parameters
    ----------
    performance: partitura.performance.PerformanceLike
        Performance information, can be a ppart, performance
    score_part: partitura.score.ScoreLike
        Score information, can be a part, score
    alignment : list
        list of alignment dicts
    """

    # nf = pt.musicanalysis.compute_note_array(score_part,
    #                                              include_pitch_spelling=True,
    #                                              include_key_signature=True,
    #                                              include_time_signature=True,
    #                                              include_metrical_position=True,
    #                                              include_grace_notes=True,
    #                                              feature_functions="all",
    #                                              force_fixed_size=True)

    nf = score_part.note_array()

    pf = pt.musicanalysis.make_performance_features(
        score_part, performance, alignment, feature_functions="all"
    )

    pf = rfn.rename_fields(
        pf, dict(zip(["p_onset", "p_duration"], ["onset_sec", "duration_sec"]))
    )
    score_fields = ["onset_beat", "duration_beat", "pitch", "id", "voice"]
    performance_fields = [
        "id",
        "onset_sec",
        "duration_sec",
        "beat_period",
        "timing",
        "articulation_log",
        "velocity",
        "pedal_feature.onset_value",
    ]
    score_f = nf[score_fields]
    performance_f = pf[performance_fields]
    merged_ = np.lib.recfunctions.join_by(
        "id", score_f, performance_f, jointype="inner"
    ).data
    return merged_


def compute_onsetwise_snote_pnote_array(
    merged_note_array, attribute_list=["velocity", "beat_period", "articulation_log"]
):
    unique_onset_idx = pt.musicanalysis.performance_codec.get_unique_onset_idxs(
        merged_note_array["onset_beat"]
    )
    onset_val = pt.musicanalysis.performance_codec.notewise_to_onsetwise(
        merged_note_array["onset_sec"], unique_onset_idx
    )
    onset_beat_val = pt.musicanalysis.performance_codec.notewise_to_onsetwise(
        merged_note_array["onset_beat"], unique_onset_idx
    )
    array_collector = [
        np.array(list(zip(onset_val)), dtype=np.dtype([("onset_sec", "f4")])),
        np.array(list(zip(onset_beat_val)), dtype=np.dtype([("onset_beat", "f4")])),
    ]
    for attribute in attribute_list:
        attribute_val = pt.musicanalysis.performance_codec.notewise_to_onsetwise(
            merged_note_array[attribute], unique_onset_idx
        )
        array_collector.append(
            np.array(list(zip(attribute_val)), dtype=np.dtype([(attribute, "f4")]))
        )
    return rfn.merge_arrays(array_collector, flatten=True, usemask=False)


def save_expression_features_for_sonic_visualiser(
    merged_note_array, out_dir=".", notewise=False, onsetwise=True, beatwise=0
):
    """
    Saves a note array attribute array to a Sonic Visualiser time values file.

    Parameters
    ----------
    merged_note_array: ndarray
        note array with p and s features: use compute_snote_pnote_array

    """
    out_dir = Path(out_dir)

    if notewise:
        save_attribute_for_sonic_visualiser_time_values(
            merged_note_array, "velocity", out_dir / Path("nw_velocity.csv")
        )
        save_attribute_for_sonic_visualiser_time_values(
            merged_note_array, "beat_period", out_dir / Path("nw_beat_period.csv")
        )
        save_attribute_for_sonic_visualiser_time_values(
            merged_note_array, "timing", out_dir / Path("nw_timing.csv")
        )
        save_attribute_for_sonic_visualiser_time_values(
            merged_note_array,
            "articulation_log",
            out_dir / Path("nw_articulation_log.csv"),
        )
        save_attribute_for_sonic_visualiser_time_values(
            merged_note_array,
            "pedal_feature.onset_value",
            out_dir / Path("nw_pedal.csv"),
        )

    if onsetwise:
        onset_merged_array = compute_onsetwise_snote_pnote_array(merged_note_array)
        save_attribute_for_sonic_visualiser_time_values(
            onset_merged_array, "velocity", out_dir / Path("ow_velocity.csv")
        )
        save_attribute_for_sonic_visualiser_time_values(
            onset_merged_array, "beat_period", out_dir / Path("ow_beat_period.csv")
        )
        save_attribute_for_sonic_visualiser_time_values(
            onset_merged_array,
            "articulation_log",
            out_dir / Path("ow_articulation_log.csv"),
        )

    if beatwise:
        onset_merged_array = compute_onsetwise_snote_pnote_array(merged_note_array)
        stime_to_ptime_map = stime_to_ptime_map = interp1d(
            y=onset_merged_array["onset_sec"],
            x=onset_merged_array["onset_beat"],
            bounds_error=False,
            fill_value="extrapolate",
        )
        max_beat = np.max(onset_merged_array["onset_beat"])
        beats = np.arange(0, max_beat + beatwise, beatwise)
        perf_beats = stime_to_ptime_map(beats)
        beat_period = np.diff(perf_beats) / beatwise

        perf_beats_array = np.array(
            list(zip(perf_beats, beats)),
            dtype=np.dtype([("onset_sec", "f4"), ("onset_beat", "f4")]),
        )
        # perf_beats_array = np.array([perf_beats, beats], dtype = [("onset_sec", "f4"), ("onset_beat", "f4")])
        beat_period_array = np.array(
            list(zip(perf_beats[:-1], beat_period)),
            dtype=np.dtype([("onset_sec", "f4"), ("bw_beat_period", "f4")]),
        )
        # beat_period_array = np.array([perf_beats[:-1], beat_period], dtype = [("onset_sec", "f4"), ("bw_beat_period", "f4")])
        save_attribute_for_sonic_visualiser_time_values(
            beat_period_array, "bw_beat_period", out_dir / Path("bw_beat_period.csv")
        )
        save_attribute_for_sonic_visualiser_instants(
            perf_beats_array, "onset_beat", out_dir / Path("beats.csv")
        )


def save_sonic_visualizer_csvs(
    performance,
    score_part,
    alignment,
    out_dir=".",
    notewise=False,
    onsetwise=True,
    beatwise=0,
):
    """
    save expression features for sonic visualizer

    Parameters
    ----------
    performance : object
        a performance object
    score_part: object
        a score part object
    alignment: list
        note alignment list of dicts
    """
    out_dir = Path(out_dir)
    merged_note_array = compute_snote_pnote_array(performance, score_part, alignment)

    save_expression_features_for_sonic_visualiser(
        merged_note_array,
        out_dir=out_dir,
        notewise=notewise,
        onsetwise=onsetwise,
        beatwise=beatwise,
    )

    controls_list = [
        (cc["time"], cc["value"])
        for cc in performance.performedparts[0].controls
        if cc["number"] == 64
    ]
    controls_array = np.array(
        controls_list, dtype=np.dtype([("onset_sec", "f4"), ("pedal", "f4")])
    )
    save_attribute_for_sonic_visualiser_time_values(
        controls_array, "pedal", out_dir / Path("pedal.csv")
    )

    save_notes_for_sonic_visualiser(merged_note_array, out_dir / Path("notes.csv"))



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
This module contains pairwise distance metrics and other DP helpers.
"""
import numpy as np
from numba import jit


def element_of_metric(vec1, vec2):
    """
    metric that evaluates occurence of vec2 (scalar) in vec1 (vector n-dim)
    """
    return 1 - np.sum(vec2 == vec1)


def element_of_set_metric(element_, set_):
    """
    metric that evaluates occurence of an element in a set
    """
    if element_ in set_:
        return 0.0
    else:
        return 1.0


def element_of_set_metric_se(set_, element_):
    """
    metric that evaluates occurence of an element in a set
    """
    if element_ in set_:
        return 0.0
    else:
        return 1.0


def l2(vec1, vec2):
    """
    l2 metric between vec1 and vec2
    """
    return np.sqrt(np.sum((vec2 - vec1) ** 2))


def invert_matrix(S, inversion="reciprocal", positive=False):
    """
    simple converter from similarity to distance matrix
    and vice versa
    """
    if inversion == "reciprocal":
        D = 1 / (S + 1e-4)
    else:
        D = -S
    if positive:
        D -= D.min()
    return D


def dnw(vec1, vec2):
    """
    normalized and weighted distance for LNCO/LNSO features.
    https://ieeexplore.ieee.org/abstract/document/6333860/
    """
    vec1_l1 = np.abs(vec1).sum()
    vec2_l1 = np.abs(vec2).sum()
    dn = np.abs(vec1 - vec2).sum() / (vec1_l1 + vec2_l1 + 1e-7)  # L1 okay?
    dampening_factor = ((vec1_l1 + vec2_l1 + 1e-7) / 2) ** (
        1 / 4
    )  # safety eps not necessary
    return dn * dampening_factor


def cdist_local(arr1, arr2, metric):
    """
    compute array of pairwise distances between
    the elements of two arrays given a metric

    Parameters
    ----------
    arr1: numpy nd array

    arr2: numpy nd array

    metric: callable
        a metric function

    Returns
    -------

    pdist_array: numpy 2d array
        array of pairwise distances
    """
    l1 = len(arr1)
    l2 = len(arr2)
    pdist_array = np.ones((l1, l2)) * np.inf
    for i in range(l1):
        for j in range(l2):
            pdist_array[i, j] = metric(arr1[i], arr2[j])
    return pdist_array


@jit(nopython=True)
def bounded_recursion(prev_val, min_val=0, max_val=10, slope_at_min=1):
    """
    a recursive function which when starting at min_val,
    grows to slope_at_min after one step,
    then continues to grow and asymptotically reaches max_val
    """
    dv = max_val - min_val
    exponential_decay_factor = (slope_at_min) / dv
    normal_prev_val = (prev_val - min_val - dv) / (-dv)
    normal_next_val = (1 - exponential_decay_factor) * normal_prev_val
    next_val = -dv * normal_next_val + dv + min_val
    return next_val


def tempo_and_pitch_metric(
    pitch_set_s,
    pitch_p,
    onset_s,
    onset_p,
    prev_onset_s,
    prev_onset_p,
    tempo,  # sec / beat
    time_weight=0.5,
    tempo_factor=0.5,
):
    """
    metric that combines
    1) pitch set metric
    2) onset deviation based on tempo

    returns the distance and the tempo associated with it
    """
    # pitch stuff
    if pitch_p in pitch_set_s:
        pitch_dist = 0
    else:
        pitch_dist = 1

    # onset stuff
    estimated_onset = prev_onset_p + (onset_s - prev_onset_s) * tempo
    onset_dist = abs(onset_p - estimated_onset) / tempo  # normalize offset by tempo4

    # tempo stuff
    if onset_s - prev_onset_s > 0:
        current_tempo = (onset_p - prev_onset_p) / (onset_s - prev_onset_s)
    else:
        current_tempo = tempo + (onset_p - prev_onset_p)
    exponential_average_tempo = (
        tempo_factor * current_tempo + (1 - tempo_factor) * tempo
    )

    dist = pitch_dist + time_weight * onset_dist
    return dist, exponential_average_tempo


def onset_pitch_duration_metric(
    pitch_s,
    pitch_p,
    onset_s,
    onset_p,
    prev_onset_s,
    prev_onset_p,
    duration_s,
    duration_p,
    tempo,  # sec / beat
    weights=np.array([1, 1, 1]),  # onset, dur, pitch
    tempo_factor=0.5,
):
    """
    metric that combines
    1) pitch set metric
    2) onset deviation based on tempo
    3) duration

    returns the distance and the tempo associated with it
    """
    # pitch stuff
    if pitch_p == pitch_s:
        pitch_dist = 0
    else:
        pitch_dist = 1

    # onset stuff
    estimated_onset = prev_onset_p + (onset_s - prev_onset_s) * tempo
    onset_dist = abs(onset_p - estimated_onset) / tempo  # normalize offset by tempo4

    # duration stuff
    estimated_duration = duration_s * tempo
    duration_dist = abs(
        np.log(duration_p / estimated_duration)
    )  # abs log of duration ratio

    # tempo stuff
    if onset_s - prev_onset_s > 0:
        current_tempo = (onset_p - prev_onset_p) / (onset_s - prev_onset_s)
    else:
        current_tempo = tempo + (onset_p - prev_onset_p)
    exponential_average_tempo = (
        tempo_factor * current_tempo + (1 - tempo_factor) * tempo
    )

    dist = weights.dot(np.array([onset_dist, duration_dist, pitch_dist]))
    return dist, exponential_average_tempo



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
This module contains methods to align subparts of a performance.
"""
import numpy as np
from ..dp.nwtw import SubPartDynamicProgramming
import partitura as pt


class SubPartMatcher(object):
    """
    Subpart Alignment
    """

    def __init__(self):
        self.DP = SubPartDynamicProgramming()

    def preprocess_score(self, score):
        bass_line = score.parts[-1]
        return pt.score.Score(bass_line)

    def preprocess_performance(self, pna, sna):
        max_score_pitch = np.max(sna["pitch"])
        min_score_pitch = np.min(sna["pitch"])
        pitch_mask = np.all(
            (pna["pitch"] <= max_score_pitch, pna["pitch"] >= min_score_pitch), axis=0
        )
        return pna[pitch_mask], pna[~pitch_mask]

    def align_from_path(self, path, sna, pna, pna_top_voice=None):
        alignment = list()
        used_pidx = set()
        for snote_idx in range(len(sna)):
            pnote_idx = path[snote_idx]
            pnote = pna[pnote_idx]
            snote = sna[snote_idx]
            if np.max(pnote["pitch"] == snote["pitch"]):
                alignment.append(
                    {
                        "label": "match",
                        "score_id": str(snote["id"]),
                        "performance_id": str(pnote["id"]),
                    }
                )
                used_pidx.add(pnote["id"])
            else:
                alignment.append({"label": "deletion", "score_id": str(snote["id"])})

        for pnote in pna:
            if pnote["id"] not in used_pidx:
                alignment.append(
                    {"label": "insertion", "performance_id": str(pnote["id"])}
                )

        if pna_top_voice is not None:
            for pnote in pna_top_voice:
                alignment.append(
                    {"label": "insertion", "performance_id": str(pnote["id"])}
                )

        return alignment

    def from_note_arrays(self, sna, pna_original, preprocess_pna=True):
        """
        compute subpart alignment from note arrays.

        Parameters
        ----------
        sna: ndarray
            score note array of the subpart
        pna : ndarray
            performance note array of the full performance
        preprocess_pna: bool
            trim the performance pitch range to the subpart
            pitch range before alignment, defaults to True.
        """
        if preprocess_pna:
            pna_bottom_voice, pna_top_voice = self.preprocess_performance(
                pna_original, sna
            )
            _, path, _, _ = self.DP(pna_bottom_voice, sna)
            alignment = self.align_from_path(path, sna, pna_bottom_voice, pna_top_voice)
        else:
            _, path, _, _ = self.DP(pna_original, sna)
            alignment = self.align_from_path(path, sna, pna_original)
        return alignment

    def __call__(self, score_path, performance_path, preprocess_pna=True):
        """
        compute subpart alignment from score
        and performance paths.

        Parameters
        ----------
        score_path: str
            path to a score
        performance_path : str
            path to a performance
        preprocess_pna: bool
            trim the performance pitch range to the subpart
            pitch range before alignment, defaults to True.
        """
        score = pt.load_musicxml(score_path, force_note_ids=True)
        # re
        bscore = self.preprocess_score(score)
        sna = bscore.note_array()

        performance = pt.load_performance_midi(performance_path)
        pna_original = performance.note_array()

        alignment = self.from_note_arrays(sna, pna_original, preprocess_pna)
        return alignment



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
This module contains methods for generalized note alignment.
e.g.:
- matching a performance to a score without knowing the repeat structure 
- matching a monophonic voice to a full performance
"""

from .repeat_identification import RepeatIdentifier
from .subpart import SubPartMatcher



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
This module contains methods to visualize alignments
"""

import numpy as np
import partitura as pt
import matplotlib.pyplot as plt
import random


def plot_alignment(
    ppart_na,
    part_na,
    alignment,
    save_file=False,
    fname="note_alignment",
    random_color=False,
):
    first_note_midi = np.min(ppart_na["onset_sec"])
    last_note_midi = np.max(ppart_na["onset_sec"] + ppart_na["duration_sec"])
    first_note_start = np.min(part_na["onset_beat"])
    last_note_start = np.max(part_na["onset_beat"])
    length_of_midi = last_note_midi - first_note_midi
    length_of_xml = last_note_start - first_note_start

    length_of_pianorolls = max(10000, int(length_of_xml * 8))
    time_div_midi = int(np.floor(length_of_pianorolls / length_of_midi))
    time_div_xml = int(np.floor(length_of_pianorolls / length_of_xml))

    midi_piano_roll, perfidx = pt.utils.compute_pianoroll(
        ppart_na,
        time_unit="sec",
        time_div=time_div_midi,
        return_idxs=True,
        remove_drums=False,
    )
    xml_piano_roll, scoreidx = pt.utils.compute_pianoroll(
        part_na,
        time_unit="beat",
        time_div=time_div_xml,
        return_idxs=True,
        remove_drums=False,
    )

    plot_array = np.zeros((128 * 2 + 50, length_of_pianorolls + 800))
    dense_midi_pr = midi_piano_roll.todense()  # [:,:time_size]
    dense_midi_pr[dense_midi_pr > 0] = 1
    plot_array[:128, : xml_piano_roll.shape[1]] = xml_piano_roll.todense()
    plot_array[50 + 128 :, : midi_piano_roll.shape[1]] = dense_midi_pr

    f, axs = plt.subplots(1, 1, figsize=(100, 10))
    axs.matshow(plot_array, aspect="auto", origin="lower")
    hexadecimal_alphabets = "0123456789ABCDEF"

    if random_color:
        colors = [
            "#" + "".join([random.choice(hexadecimal_alphabets) for j in range(6)])
            for i in range(40)
        ]
    else:
        colors = ["#00FF00" for i in range(40)]

    score_dict = dict()
    for note, pos in zip(part_na, scoreidx):
        score_dict[note["id"]] = pos
    perf_dict = dict()
    for note, pos in zip(ppart_na, perfidx):
        perf_dict[note["id"]] = pos
    for i, line in enumerate(alignment):
        if line["label"] == "match":
            perf_pos = perf_dict[line["performance_id"]]
            score_pos = score_dict[line["score_id"]]
            axs.plot(
                [score_pos[1], perf_pos[1]],
                [score_pos[0], 128 + 50 + perf_pos[0]],
                "o-",
                lw=2,
                c=colors[i % 40],
            )

    if save_file:
        plt.savefig(fname + ".png")
        plt.close(f)
    else:
        plt.show()


def plot_alignment_comparison(
    ppart_na,
    part_na,
    alignment1,
    alignment2,
    save_file=False,
    return_figure=False,
    figsize=(100, 10),
    fname="note_alignments",
):
    first_note_midi = np.min(ppart_na["onset_sec"])
    last_note_midi = np.max(ppart_na["onset_sec"] + ppart_na["duration_sec"])
    first_note_start = np.min(part_na["onset_beat"])
    last_note_start = np.max(part_na["onset_beat"])
    length_of_midi = last_note_midi - first_note_midi
    length_of_xml = last_note_start - first_note_start

    length_of_pianorolls = max(10000, int(length_of_xml * 8))
    time_div_midi = int(np.floor(length_of_pianorolls / length_of_midi))
    time_div_xml = int(np.floor(length_of_pianorolls / length_of_xml))

    midi_piano_roll, perfidx = pt.utils.compute_pianoroll(
        ppart_na,
        time_unit="sec",
        time_div=time_div_midi,
        return_idxs=True,
        remove_drums=False,
    )
    xml_piano_roll, scoreidx = pt.utils.compute_pianoroll(
        part_na,
        time_unit="beat",
        time_div=time_div_xml,
        return_idxs=True,
        remove_drums=False,
    )

    plot_array = np.zeros((128 * 2 + 50, length_of_pianorolls + 800))
    dense_midi_pr = midi_piano_roll.todense()  # [:,:time_size]
    dense_midi_pr[dense_midi_pr > 0] = 1
    plot_array[:128, : xml_piano_roll.shape[1]] = xml_piano_roll.todense()
    plot_array[50 + 128 :, : midi_piano_roll.shape[1]] = dense_midi_pr

    f, axs = plt.subplots(1, 1, figsize=figsize)
    axs.matshow(plot_array, aspect="auto", origin="lower")
    hexadecimal_alphabets = "0123456789ABCDEF"

    colors1 = ["#00FF00" for i in range(40)]
    colors2 = ["#0000FF" for i in range(40)]
    colors3 = ["#FF0000" for i in range(40)]
    colors4 = ["#FF00FF" for i in range(40)]

    n1_but_not_n2 = [al for al in alignment1 if not al in alignment2]
    n2_but_not_n1 = [al for al in alignment2 if not al in alignment1]

    score_dict = dict()
    for note, pos in zip(part_na, scoreidx):
        score_dict[note["id"]] = pos
    perf_dict = dict()
    for note, pos in zip(ppart_na, perfidx):
        perf_dict[note["id"]] = pos
    for i, line in enumerate(alignment1):
        if line["label"] == "match":
            perf_pos = perf_dict[line["performance_id"]]
            score_pos = score_dict[line["score_id"]]
            axs.plot(
                [score_pos[1], perf_pos[1]],
                [score_pos[0], 128 + 50 + perf_pos[0]],
                "o-",
                lw=2,
                c=colors1[i % 40],
            )
    for i, line in enumerate(alignment2):
        if line["label"] == "match":
            perf_pos = perf_dict[line["performance_id"]]
            score_pos = score_dict[line["score_id"]]
            axs.plot(
                [score_pos[1], perf_pos[1]],
                [score_pos[0], 128 + 50 + perf_pos[0]],
                "o-",
                lw=2,
                c=colors2[i % 40],
            )
    for i, line in enumerate(n1_but_not_n2):
        if line["label"] == "match":
            perf_pos = perf_dict[line["performance_id"]]
            score_pos = score_dict[line["score_id"]]
            axs.plot(
                [score_pos[1], perf_pos[1]],
                [score_pos[0], 128 + 50 + perf_pos[0]],
                "o-",
                lw=2,
                c=colors3[i % 40],
            )
    for i, line in enumerate(n2_but_not_n1):
        if line["label"] == "match":
            perf_pos = perf_dict[line["performance_id"]]
            score_pos = score_dict[line["score_id"]]
            axs.plot(
                [score_pos[1], perf_pos[1]],
                [score_pos[0], 128 + 50 + perf_pos[0]],
                "o-",
                lw=2,
                c=colors4[i % 40],
            )

    if save_file:
        plt.savefig(fname + ".png")
        plt.close(f)
    else:
        if return_figure:
            return f
        else:
            plt.show()


def plot_alignment_mappings(
    ppart_na,
    part_na,
    score_to_performance_mapping1,
    score_to_performance_mapping2,
    save_file=False,
    fname="onset_alignments",
):
    first_note_midi = np.min(ppart_na["onset_sec"])
    last_note_midi = np.max(ppart_na["onset_sec"] + ppart_na["duration_sec"])
    first_note_start = np.min(part_na["onset_beat"])
    last_note_start = np.max(part_na["onset_beat"])
    length_of_midi = last_note_midi - first_note_midi
    length_of_xml = last_note_start - first_note_start

    length_of_pianorolls = max(10000, int(length_of_xml * 8))
    time_div_midi = int(np.floor(length_of_pianorolls / length_of_midi))
    time_div_xml = int(np.floor(length_of_pianorolls / length_of_xml))

    midi_piano_roll, perfidx = pt.utils.compute_pianoroll(
        ppart_na,
        time_unit="sec",
        time_div=time_div_midi,
        return_idxs=True,
        remove_drums=False,
    )
    xml_piano_roll, scoreidx = pt.utils.compute_pianoroll(
        part_na,
        time_unit="beat",
        time_div=time_div_xml,
        return_idxs=True,
        remove_drums=False,
    )

    midi_to_pr_mapping = lambda x: int((x - first_note_midi) * time_div_midi)
    xml_to_pr_mapping = lambda x: int((x - first_note_start) * time_div_xml)

    plot_array = np.zeros((128 * 2 + 50, length_of_pianorolls + 800))
    dense_midi_pr = midi_piano_roll.todense()  # [:,:time_size]
    dense_midi_pr[dense_midi_pr > 0] = 1
    plot_array[:128, : xml_piano_roll.shape[1]] = xml_piano_roll.todense()
    plot_array[50 + 128 :, : midi_piano_roll.shape[1]] = dense_midi_pr

    f, axs = plt.subplots(1, 1, figsize=(100, 10))
    axs.matshow(plot_array, aspect="auto", origin="lower")
    for unique_onset in np.unique(part_na["onset_beat"]):
        perf_pos1 = midi_to_pr_mapping(score_to_performance_mapping1(unique_onset))
        perf_pos2 = midi_to_pr_mapping(score_to_performance_mapping2(unique_onset))
        score_pos = xml_to_pr_mapping(unique_onset)
        axs.plot([score_pos, perf_pos1], [128, 128 + 50], "o-", lw=2, c="#00FF00")
        axs.plot([score_pos, perf_pos2], [128, 128 + 50], "o-", lw=2, c="#0000FF")

    if save_file:
        plt.savefig(fname + ".png")
        plt.close(f)
    else:
        plt.show()



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
This module contains methods to evaluate
- note matching via f scores for match, insertion, and deletion
- score following / temporal alignment via asynchrony
"""
from typing import List
import numpy as np
import partitura as pt
import os


def fscore_alignments(
    prediction: List[dict],
    ground_truth: List[dict],
    types: List[str],
    return_numbers=False,
) -> (float, float, float):
    """
    Parameters
    ----------
    prediction: List of dictionaries containing the predicted alignments
    ground_truth: List of dictionaries containing the ground truth alignments
    types: List of alignment types to consider for evaluation (e.g ['match', 'deletion', 'insertion']

    Returns
    -------
    precision, recall, f score
    """

    pred_filtered = list(filter(lambda x: x["label"] in types, prediction))
    gt_filtered = list(filter(lambda x: x["label"] in types, ground_truth))

    filtered_correct = [pred for pred in pred_filtered if pred in gt_filtered]

    n_pred_filtered = len(pred_filtered)
    n_gt_filtered = len(gt_filtered)
    n_correct = len(filtered_correct)

    if n_pred_filtered > 0 or n_gt_filtered > 0:
        precision = n_correct / n_pred_filtered if n_pred_filtered > 0 else 0.0
        recall = n_correct / n_gt_filtered if n_gt_filtered > 0 else 0
        f_score = (
            2 * precision * recall / (precision + recall)
            if (precision + recall) > 0
            else 0
        )
    else:
        # no prediction and no ground truth for a given type -> correct alignment
        precision, recall, f_score = 1.0, 1.0, 1.0

    if return_numbers:
        return precision, recall, f_score, len(pred_filtered), len(gt_filtered)
    else:
        return precision, recall, f_score


def print_fscore_alignments(prediction: List[dict], ground_truth: List[dict]):
    print("------------------")
    types = ["match", "insertion", "deletion"]
    for alignment_type in types:
        precision, recall, f_score = fscore_alignments(
            prediction, ground_truth, alignment_type
        )
        print("Evaluate ", alignment_type)
        print(
            "Precision: ",
            format(precision, ".3f"),
            "Recall ",
            format(recall, ".3f"),
            "F-Score ",
            format(f_score, ".3f"),
        )
        print("------------------")


def evaluate_asynchrony(target_ponsets, tracked_ponsets):
    asynchrony = target_ponsets - tracked_ponsets
    abs_asynch = abs(asynchrony)
    mean_asynch = np.median(abs_asynch)
    lt_25ms = np.mean(abs_asynch <= 0.025)
    lt_50ms = np.mean(abs_asynch <= 0.05)
    lt_100ms = np.mean(abs_asynch <= 0.1)
    return mean_asynch, lt_25ms, lt_50ms, lt_100ms


def evaluate_score_following(
    performance_note_array=None,
    score_note_array=None,
    gt_alignment=None,
    alignment=None,
    out_dir="",
    file_suffix="",
    write_to_file=False,
    print_results=False,
):
    """
    Parameters
    ----------
    performance_note_array: np.ndarray
    score_note_array: np.ndarray
    gt_alignment: List[dict]
    alignment: List[dict]
    out_dir: str
    print_results: bool

    Returns
    -------
    mean_asynch: float
    lt_25ms: float
    lt_50ms: float
    lt_100ms: float

    """

    (
        _,
        stime_to_ptime_map_gt,
    ) = pt.musicanalysis.performance_codec.get_time_maps_from_alignment(
        performance_note_array, score_note_array, gt_alignment, remove_ornaments=True
    )

    (
        _,
        stime_to_ptime_map,
    ) = pt.musicanalysis.performance_codec.get_time_maps_from_alignment(
        performance_note_array, score_note_array, alignment, remove_ornaments=True
    )
    sid1 = set()
    sid2 = set()
    for al in alignment:
        # Get only matched notes (i.e., ignore inserted or deleted notes)
        if al["label"] == "match":
            sid1.add(al["score_id"])
    for al in gt_alignment:
        if al["label"] == "match":
            sid2.add(al["score_id"])

    all_sid = list(sid1.intersection(sid2))

    tracked_sonsets = np.array(
        [
            score_note["onset_beat"]
            for score_note in score_note_array
            if score_note["id"] in all_sid
        ]
    )
    tracked_sonsets = np.unique(tracked_sonsets)

    target_ponsets = stime_to_ptime_map_gt(tracked_sonsets)
    tracked_ponsets = stime_to_ptime_map(tracked_sonsets)

    mean_asynch, lt_25ms, lt_50ms, lt_100ms = evaluate_asynchrony(
        target_ponsets=target_ponsets,
        tracked_ponsets=tracked_ponsets,
    )

    if write_to_file:
        results_fn = os.path.join(out_dir, file_suffix + "results.csv")
        with open(results_fn, "w") as f:
            f.write("mean_asynch_ms,leq_25ms_%,leq_50ms_%,leq_100ms_%\n")

            if print_results:
                print(
                    f"Mean asynchrony (ms): {mean_asynch * 1000:.2f}\nAsynchrony <= 25ms (%): {lt_25ms * 100:.1f}\n"
                    f"Asynchrony <= 50ms (%): {lt_50ms * 100:.1f}\nAsynchrony <= 100ms (%): {lt_100ms * 100:.1f}\n"
                )
            f.write(
                f"{mean_asynch * 1000:.2f},{lt_25ms * 100:.1f},"
                f"{lt_50ms * 100:.1f},{lt_100ms * 100:.1f}"
            )

    else:
        return mean_asynch, lt_25ms, lt_50ms, lt_100ms



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
This module contain methods for (repeat) structure identification
"""
import partitura as pt
import numpy as np
from ..dp.nwtw import BoundedSmithWaterman
import matplotlib.pyplot as plt


class RepeatIdentifier(object):
    """
    method wrapper to compute the most likely and musically sensible
    (starts at the start, ends at the end, repeats a valid number of times)
    sequence of score sections that correspond to an input performance.

    """

    def __init__(self, max_number_of_paths=100000):
        self.directions = np.array([[1, 1], [1, 0]])
        self.dists = np.array([1, 1])
        self.matcher = BoundedSmithWaterman(
            threshold=0.5,
            gamma_penalty=-1,
            gamma_match=1,
            directions=self.directions,
            directional_distances=self.dists,
            gain_max_val=10,
        )
        self.max_number_of_paths = max_number_of_paths

    def prepare_score(self, score):
        # score representation
        # score = pt.load_musicxml(score_path)
        part = pt.score.merge_parts(score.parts)
        score_note_array = part.note_array()
        unique_onsets = np.unique(score_note_array["onset_beat"])
        # # create pitch set representation
        score_pitches_at_onsets = list()
        for onset in unique_onsets:
            score_pitches_at_onsets.append(
                set(score_note_array[score_note_array["onset_beat"] == onset]["pitch"])
            )
        return part, unique_onsets, score_pitches_at_onsets

    def prepare_performance(self, perf):
        # perf = pt.load_performance_midi(perf_path)
        # performance representation
        perf_note_array = perf.note_array()
        perf_pitches = perf_note_array["pitch"]
        return perf_note_array, perf_pitches

    def extract_segments(self, part, unique_onsets, verbose=False):
        if verbose:
            print("*" * 20)
            print("SEGMENTS")
            print(pt.score.pretty_segments(part))
            print("*" * 20)
        # segments and paths
        pt.score.add_segments(part, force_new=True)
        segments = pt.score.get_segments(part)
        paths = pt.score.get_paths(part)
        # map segments to the score input to the alignment
        segment_onset_idx = {}
        segment_onsets = {}
        for seg_id in segments.keys():
            start = part.beat_map(segments[seg_id].start.t)
            end = part.beat_map(segments[seg_id].end.t)
            onset_mask = np.where(
                np.all((unique_onsets >= start, unique_onsets < end), axis=0)
            )
            segment_onset_idx[seg_id] = np.arange(len(unique_onsets))[onset_mask[0]]
            segment_onsets[seg_id] = unique_onsets[onset_mask[0]]

        if len(paths) > self.max_number_of_paths:
            pt.score.add_segments(part, force_new=True)
            max_paths_w_leap = pt.score.get_paths(
                part, no_repeats=False, all_repeats=True, ignore_leap_info=False
            )
            pt.score.add_segments(part, force_new=True)
            max_paths_wt_leap = pt.score.get_paths(
                part, no_repeats=False, all_repeats=True, ignore_leap_info=True
            )
            pt.score.add_segments(part, force_new=True)
            min_paths = pt.score.get_paths(
                part, no_repeats=True, all_repeats=False, ignore_leap_info=True
            )

            paths = max_paths_w_leap + max_paths_wt_leap + min_paths
        return paths, segment_onset_idx, segment_onsets

    def partial_backtrack(
        self,
        B,
        cost,
        starting_n,
        starting_m,
        ending_n,
        ending_m,
        directions=np.array([[1, 0], [1, 1], [0, 1]]),
        unmatched_cost=-10,
        matched_cost=10,
    ):
        n = ending_n  # N - 1
        m = ending_m  # M - 1
        step = [m, n]
        path = [step]
        final_cost = cost[m, n]
        current_cost = cost[m, n]
        # initialize boolean variables for stopping decoding
        costs = [current_cost]
        n_start_below_1 = 0
        close_to_start = 10
        while (
            (n > max(starting_n - 1, 0))
            and (m > max(starting_m - 1, 0))
            and (current_cost > 0)
            and not (abs(n - starting_n) < close_to_start and n_start_below_1 > 2)
        ):  # stop iterating if cost is low close to the start
            backtracking_pointer = B[m, n]
            bt_vector = directions[backtracking_pointer]
            m -= bt_vector[0]
            n -= bt_vector[1]
            step = [m, n]
            # append next step to the path
            path.append(step)
            # update current cost
            current_cost = cost[m, n]
            costs.append(current_cost)
            if current_cost < 1:
                n_start_below_1 += 1
            else:
                n_start_below_1 = 0

        # remove path appendix of cost below 1 if it is in the beginning
        if n_start_below_1 > 0:
            path = path[:-n_start_below_1]

        # check if the path roughly covers the segment
        if abs(n - starting_n) > close_to_start or len(path) <= 1:
            output_path = None
            path_cost = unmatched_cost

        elif len(path) > 10 and max(costs) > 7:
            output_path = np.array(path, dtype=np.int32)[
                -2::-1
            ]  # remove last path element
            path_cost = matched_cost

        elif len(path) <= 10:
            output_path = np.array(path, dtype=np.int32)[
                -2::-1
            ]  # remove last path element
            path_cost = matched_cost

        else:
            output_path = None
            path_cost = unmatched_cost

        return output_path, path_cost

    def compute_path_gain(
        self, cost, path, backtracking, segment_onset_idx, directions
    ):
        ending_m = cost.shape[0] - 1  # the end of the performance
        starting_m = 0
        path_gain = 0
        full_path = np.array([[cost.shape[0], cost.shape[1]]])
        full_path_list = list()
        for seg_id in path.path[::-1]:
            starting_n = segment_onset_idx[seg_id][0]
            ending_n = segment_onset_idx[seg_id][-1]
            partial_path, partial_gain = self.partial_backtrack(
                backtracking,
                cost,
                starting_n,
                starting_m,
                ending_n,
                ending_m,
                directions,
                unmatched_cost=-10,
            )
            path_gain += partial_gain
            if partial_path is not None:
                ending_m = partial_path[0, 0] - 1
                full_path = np.row_stack((partial_path, full_path))
                full_path_list.append((seg_id, partial_path))

        return path_gain, full_path, full_path_list[::-1]

    def __call__(self, score, performance, verbose=False, plot=False):
        """
        Parameters
        ----------
        score: object
            a score object
        performance : object
            a performance object
        """

        part, unique_onsets, score_pitches_at_onsets = self.prepare_score(score)
        perf_note_array, perf_pitches = self.prepare_performance(performance)
        # compute coste
        cost, backtracking = self.matcher(perf_pitches, score_pitches_at_onsets)

        # figure out the possible score segments
        paths, segment_onset_idx, segment_onsets = self.extract_segments(
            part, unique_onsets, verbose=verbose
        )

        if len(paths) < 2:
            print("no structural variations!")
            print("*" * 20)
            return None

        path_gains = {}
        for path in paths:
            path_string = "".join(path.path)
            if verbose:
                print("Testing path:", path_string)
            path_gain, full_path, full_path_list = self.compute_path_gain(
                cost, path, backtracking, segment_onset_idx, directions=self.directions
            )
            path_gains[path_gain] = (path_string, path)

        max_gain = max([k for k in path_gains.keys()])
        found_path, found_path_object = path_gains[max_gain]
        if verbose:
            print("best fitting path: ", found_path)

        if plot:
            colors = ["r", "g", "b"]
            plt.imshow(cost, aspect="auto")
            for pp_no, pp_val in enumerate(full_path_list):
                partial_path_id, partial_path = pp_val
                plt.plot(partial_path[:, 1], partial_path[:, 0], c=colors[pp_no % 3])
                plt.text(
                    partial_path[0, 1],
                    partial_path[0, 0],
                    partial_path_id,
                    c=colors[pp_no % 3],
                    fontsize=12,
                )
            plt.savefig(performance + "_" + path_string + "_.png")
            plt.close()
        path_gains[path_gain] = path_string

        return found_path, found_path_object



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
This module contains methods to visualize, export, and
evaluate aligned symbolic music data.
"""

from .io import save_piano_precision_csv, save_sonic_visualizer_csvs
from .eval import (
    fscore_alignments,
    print_fscore_alignments,
    evaluate_asynchrony,
    evaluate_score_following,
)
from .plot import plot_alignment, plot_alignment_comparison, plot_alignment_mappings



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This module includes tests for alignment utilities.
"""
import unittest
import numpy as np
from parangonar import RepeatIdentifier, SubPartMatcher, fscore_alignments
import partitura as pt
from copy import copy

RNG = np.random.RandomState(1984)
from tests import MATCH_FILES, P_FILES, S_FILES


class TestMismatch(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.perf_match, cls.alignment, cls.score_match = pt.load_match(
            filename=MATCH_FILES[0], create_score=True
        )
        cls.perf_mid = pt.load_performance_midi(P_FILES[0])
        cls.score_xml = pt.load_musicxml(S_FILES[0])

    def test_repeat_identification(self, **kwargs):
        ri = RepeatIdentifier()
        found_path, _ = ri(self.score_xml, self.perf_mid)
        true_path = "AABBCCDEDFAB"
        self.assertTrue(true_path == found_path)

    def test_subpart_alignment(self, **kwargs):
        # create a subpart
        sna = self.score_match.note_array()
        sna_mask = np.arange(0, len(sna), 12)
        sna_subpart = sna[sna_mask]
        sort_idx = np.argsort(sna_subpart["onset_beat"])
        sna_subpart = sna_subpart[sort_idx]
        new_alignment = list()
        for m in self.alignment:
            if m["label"] == "match":
                if m["score_id"] in sna_subpart["id"]:
                    new_alignment.append(m)
                else:
                    new_alignment.append(
                        {"label": "insertion", "performance_id": m["performance_id"]}
                    )
                    new_alignment.append(
                        {"label": "deletion", "score_id": m["score_id"]}
                    )
            else:
                new_alignment.append(m)

        pna = self.perf_match.note_array()
        matcher = SubPartMatcher()
        pred_alignment = matcher.from_note_arrays(sna_subpart, pna, True)
        _, _, f_score = fscore_alignments(pred_alignment, new_alignment, "match")

        self.assertTrue(f_score > 0.94)


if __name__ == "__main__":
    unittest.main()
